/*
  I was inspired by a similar project I saw on an Electronoobs YouTube video. It was really simple and well presented as usual.
  It demonstrates the theory of how we will go about measuring the inductance. Please have a look if you havn't seen it.
  https://www.youtube.com/watch?v=BGjV5vUvGPc&t=472s
  http://electronoobs.com/images/Arduino/tut_10/339_circuit.png

  My approach is similar as I am pulsing a resonant tank circuit and measuring the period of the resulting sinewave however,
  there are some key differences. This inductance meter uses an ATtiny85 microcontroller and the internal comparator which replaces the external LM339 comparator.
  An emitter-follower transistor is added to provide more drive current to the resonant tank circuit allowing for inductance measurements down to 1uH.
  Instead of using the pulsIn() function, I use Timer0 along with internal interrupt from the comparator creating a frequency counter that results
  in more accurate inductor measurment. Finally, a 0.96 OLED display replaces the larger LCD for a more compact design.
  
*/

// Use the Adafruit's TinyWireM library:
#include <TinyWireM.h>
#include <Tiny4kOLED.h>

unsigned int count_1;// 50us_count (0 - 65535) 0 - 3.27sec in 50us increments
int count_2;// TCNT0 Count (0-99) 0 - 50us in 0.5us increments
unsigned int count_us;// micro-seconds
unsigned int TIMR_50us_cntr;// Timer0 ISR counter
unsigned int period = 20000;// 20000 * 50us = 1sec) Display and Pulse interval
int trigger_cnt;// # of comparator output triggers detected
bool mH = false;// Change units between mH and uH

void setup() {
  oled.begin(128, 64, sizeof(tiny4koled_init_128x64br), tiny4koled_init_128x64br);
  oled.on();
  oled.clear();
  oled.setFont(FONT8X16P);
  oled.setCursor(0, 0);
  oled.print("  Inductance   ");
  oled.setCursor(0, 2);
  oled.print("     Meter     ");
 
  sei();// Enable Global Interrupts
  
  pinMode(4, OUTPUT);// Drive Pulse on PB4
  pinMode(3, OUTPUT);// Comparator output pulse on PB3

  // ATtiny85 is running on 16MHz Internal
  // Timer0 is used as the master timer for generating the puse, display handleing and a precision frequency counter
  // It operates in "CTC Mode" and counts from 0-100 and rolls over again and each increment is 500ns
  // an interrupt is generated at OCRA
  /*
    TCCR0A - Timer/Counter0 Control Register (page 79)
    COM0A1=0 COM0A0=0 COM0B1=0 COM0B0=0 (Normal port operation, OC0A/OC0B disconnected)
    WGM01=1 WGM00=0 (Sets CTC Mode along with WGM02 = 0 in the TCCR0B Register)
  */
  TCCR0A = B00000010;
  /*
    WGM02=0 (Sets CTC Mode along with WGM00 and WGM01 in the TCCR0A Register)
    CS02=0 CS01=1 CS00=0 (Divide by 8)
  */
  TCCR0B = B00000010;
  /*
    TIMSK -Timer/Counter Interrupt Mask Register
    OCIE0A=1(Timer/Counter1 Compare MatchA, interrupt is enabled)
  */
  TIMSK = B00010000;
  /*
    //TIFR – Timer/Counter Interrupt Flag Register
    //OCF0A=1(When Interrups are enabled and  OCIE1A=1 OCF1A=1 the Timer/Counter1 A compare match interrupt is executed.
  */
  TIFR = B00010000;

  OCR0A = 100;// 50us ISR
  
  // Internal Comparator
  // The internal comparator is used to detect the rising edge of a sinewave generated by the resonant tank circuit excited by a pulse
  // When Bit6 is written logic zero, AIN1 is applied to the negative input of the Analog Comparator. (PB1)
  ADCSRB &= B10111111;
  /*
    ACSR – Analog Comparator Control and Status Register (page 120)
    Comparator is using the internal 1.1V Bandgap reference applied to the AIN0 input and interrupt is enabled.
    It is compared with input AIN1 on PB1 and set to trigger the interrupt on the falling edge.
  */
  ACSR |= B01011010; // 
      //   ||||||||_ ACIS0: Analog Comparator Interrupt Mode Select works together with ACIS1
      //   |||||||__ ACIS1:ACIS0 10= falling edge; 11= raising edge; 00= toggle. Note this refers to the output of the internal comparator NOT the input signal at PB1
      //   ||||||___ Reserved
      //   |||||____ ACIE  Interrupt Enable
      //   ||||_____ ACI   Analog Comparator Interrupt Flag (R/W), write 1 to clear it.
      //   |||______ ACO   Analog Comparator Output (read only)
      //   ||_______ ACBG  1= Bandgap Select (1.1V); 0= compare with AIN0 (pin 5)
      //   |________ ACD   Analog Comparator Disable

}

// 16Mhz / 8 = 0.5us per count
// 16Mhz / 8 / 100 = 20kHz = TIMR_50us ISR
ISR(TIMER0_COMPA_vect)
{
  //Each increment of this counter represents 100 0.5us counts
  TIMR_50us_cntr++;
}

// Comparator Output triggers on each rising edge of the sinewave. Note: The ACSR register bits ACIS1:ACIS0 are set to detect the falling edge since
// this refers to the output of the internal comparator that triggers the ISR and NOT the input signal on AIN1 (PB1) (page 120 / 121 of the spec)
ISR (ANA_COMP_vect)
{
  // Capture the first two rising edges (360º sinewave)and create a single positive pulse on output D3
  if(trigger_cnt == 0)
  {
    //Set PB3 HIGH
    PORTB |= B00001000;
    // Reset counters    
    TCNT0 = 0;
    TIMR_50us_cntr = 0;
  }
  if(trigger_cnt == 1)
  {
    //Set PB3 LOW
    PORTB &= B11110111;
    // Get timer values
    count_1 = TIMR_50us_cntr;
    count_2 = TCNT0;
  } 
  trigger_cnt++;
}

void pulse() {
  // Use Timer0 to create a 50us Pulse
  // Reset Timer0 to zero
  TCNT0 = 0;

  // While Timer0 < 45us
  while(TCNT0 < 90)
  {
    // Disable Comparator Interrupt
    ACSR &= B11110111;
    //Set PB3 LOW
    PORTB &= B11110111;
    digitalWrite(4, HIGH);
  }
  digitalWrite(4, LOW);
  // Enable Comparitor Interrupt
  ACSR |= B00001000;
  trigger_cnt = 0;  
}

void loop() {
    // Calculate and display once every second
    // 50us * 20000 = 1sec
    if(TIMR_50us_cntr >= period)
    {
      // Each counting increment of Timer0 TCNT0 is 0.5us
      // count_1 = TIMR_50us_cntr meaning that every count_1 increment contains 100 Timer0 increments(50us)
      // count_2 = TCNT0 count which is 0.5us increments
      // So by adding the two counts and dividing by 2 we get count_us which is the total # of micro-seconds 
      count_us = ((count_1 * 100) + count_2)/2;

      //This is our known capacitance value
      double capacitance = 10E-6;
      // Calculate the frequency
      unsigned long freq = 1 / (count_us * 1E-6);//
      // The formula to find inductance when capacitance and frequency are known is  L = 1 / (4π^2 f^2 C)   Inductance [H]
      double inductance = 1./(capacitance*pow(freq,2)*4.*pow(M_PI,2));//
      if(inductance >= 1E-3)
      {
        mH = true;
        inductance *= 1E3;
      }
      else
      {
        mH = false;
        inductance *= 1E6;
      }
      oled.setCursor(0,4);
      oled.print("                ");      
      oled.setCursor(0,4);
      oled.print(freq);
      oled.setCursor(64,4);
      oled.print("Hz");
      oled.setCursor(0,6);
      oled.print("                ");          
      oled.setCursor(0,6);
      oled.print(inductance,1);
      oled.setCursor(64,6);
      if(mH)
        oled.print("mH");
      else
        oled.print("uH");
        
      //Reset all counters to zero
      TIMR_50us_cntr = 0;
      TCNT0 = 0;
      count_1 = 0;
      count_2 = 0;

      // Start a new pulse
      pulse(); 
    }
}
